from typing import List, Any
import os
import fiona
import fiona.transform
import numpy as np
import rasterio
import rasterio.mask
import shapely.geometry
from tqdm import tqdm
import glob
import cv2
from rasterio.features import rasterize
import matplotlib.pyplot as plt

from blueness import module
from blue_objects import objects, file
from blue_objects.metadata import get_from_object, post_to_object
from blue_objects.logger.matrix import log_matrix
from blue_objects.graphics.signature import sign_filename
from blue_geo import fullname as blue_geo_fullname
from roofai import fullname as roofai_fullname

from palisades.host import signature
from palisades import env
from palisades import NAME
from palisades import fullname
from palisades.logger import logger

NAME = module.name(__file__, NAME)


# based on https://github.com/microsoft/building-damage-assessment/blob/main/merge_with_building_footprints.py
def analyze_buildings(
    object_name: str,
    buffer: float,
    verbose: bool = False,
) -> bool:
    prediction_filename = get_from_object(
        object_name,
        "predict.output_filename",
        "",
    )
    if not prediction_filename:
        logger.error("predict.output_filename not found in metadata.")
        return False
    prediction_full_filename = objects.path_of(prediction_filename, object_name)

    datacube_id = get_from_object(
        object_name,
        "predict.datacube_id",
        "",
    )
    reference_filename = get_from_object(
        object_name,
        "predict.reference_filename",
        "",
    )
    reference_full_filename = objects.path_of(reference_filename, datacube_id)

    footprint_filename_list = glob.glob(objects.path_of("*.gpkg", object_name))
    if not footprint_filename_list:
        logger.error(f"*.gpkg not found in {object_name}.")
        return False
    footprint_filename = footprint_filename_list[0]

    logger.info(
        "{}.analyze_buildings({:.0} m): {} / {} : {}".format(
            NAME,
            buffer,
            object_name,
            prediction_filename,
            file.name_and_extension(footprint_filename),
        )
    )

    with rasterio.open(prediction_full_filename, "r") as src:
        predictions_crs = src.crs.to_string()

    with fiona.open(footprint_filename, "r") as src:
        footprints_crs = src.crs.to_string()

    # Clip building footprints to image data mask
    logger.info(f"crs: {footprints_crs} -> {predictions_crs}")
    projected_building_geoms = []
    with fiona.open(footprint_filename) as f:
        for row in tqdm(f):
            projected_geom = fiona.transform.transform_geom(
                footprints_crs, predictions_crs, row["geometry"]
            )
            projected_building_geoms.append(projected_geom)

    logger.info("analyzing {:,} building(s)".format(len(projected_building_geoms)))
    list_of_building_info: List[Any] = []

    with rasterio.open(
        prediction_full_filename,
    ) as prediction_raster, rasterio.open(
        reference_full_filename,
    ) as reference_raster:
        pixel_size = prediction_raster.res
        pixel_area = pixel_size[0] * pixel_size[1]
        logger.info(f"pixel size: {pixel_size} m = {pixel_area:.2f} m^2")

        for index, building_geom in tqdm(enumerate(projected_building_geoms)):
            building_info = {
                "id": index,
            }

            building_shape = shapely.geometry.shape(building_geom)
            building_shape_buffered = building_shape.buffer(buffer)

            prediction_mask, transform = rasterio.mask.mask(
                prediction_raster,
                [building_shape_buffered],
                crop=True,
                nodata=0,
                filled=True,
            )
            prediction_mask = prediction_mask[0].astype(np.float32) / 255
            building_info["damage"] = float(np.mean(prediction_mask))

            rasterized_shape = rasterize(
                [(building_shape, 1)],  # value 1 to all pixels inside the shape
                out_shape=prediction_mask.shape,
                transform=transform,
                fill=0,  # background
                dtype="uint8",
            )
            building_info["area"] = float(np.sum(rasterized_shape) * pixel_area)

            building_info["thumbnail"] = "thumbnail-{}{:06}.png".format(
                file.name(prediction_filename),
                index,
            )

            building_halo = (rasterized_shape.astype(np.float32) + 1) / 2

            reference_mask, transform = rasterio.mask.mask(
                reference_raster,
                [building_shape_buffered],
                crop=True,
                nodata=0,
                filled=True,
            )
            reference_mask = np.transpose(reference_mask, (1, 2, 0))
            reference_mask = (
                reference_mask.astype(np.float32) * building_halo[:, :, np.newaxis]
            )

            if not log_matrix(
                matrix=prediction_mask * building_halo,
                suffix=[reference_mask],
                header=objects.signature(
                    info=reference_filename,
                    object_name=datacube_id,
                )
                + [
                    f"{object_name} / {index:06}",
                    file.name_and_extension(footprint_filename),
                    "area: {:.1f} sq. m".format(building_info["area"]),
                    "damage: {:03.2f}%".format(
                        100 * building_info["damage"],
                    ),
                    "pixel_size: {}".format(
                        " x ".join(
                            ["{:.2f} cm".format(10 * value) for value in pixel_size]
                        )
                    ),
                    f"buffer: {buffer:.1f} m",
                ],
                footer=[
                    fullname(),
                    blue_geo_fullname(),
                    roofai_fullname(),
                ],
                colormap=cv2.COLORMAP_SPRING,
                dynamic_range=[0, 1],
                filename=objects.path_of(
                    building_info["thumbnail"],
                    object_name,
                ),
                log=verbose,
            ):
                return False

            if verbose:
                logger.info("damage: {:03.2f}%".format(100 * building_info["damage"]))

            list_of_building_info += [building_info]

    schema = {
        "geometry": "MultiPolygon",
        "properties": {
            "id": "int",
            "area": "float",
            "damage": "float",
            "thumbnail": "str",
        },
    }

    output_filename = objects.path_of(
        "analysis.gpkg",
        object_name,
    )

    logger.info("saving...")
    with fiona.open(
        output_filename,
        "w",
        driver="GPKG",
        crs=predictions_crs,
        schema=schema,
    ) as f:
        for index, (geom, building_info) in enumerate(
            tqdm(
                zip(
                    projected_building_geoms,
                    list_of_building_info,
                )
            )
        ):
            shape = shapely.geometry.shape(geom)
            if geom["type"] == "Polygon":
                geom = shapely.geometry.mapping(shapely.geometry.MultiPolygon([shape]))

            row = {
                "type": "Feature",
                "geometry": geom,
                "properties": {
                    "id": index,
                    "area": building_info["area"],
                    "damage": building_info["damage"],
                    "thumbnail": building_info["thumbnail"],
                },
            }
            f.write(row)

    logger.info(f"-> {output_filename}")

    if not post_to_object(
        object_name,
        "analysis",
        {
            "building-count": len(list_of_building_info),
            "damaged-count": len(
                [
                    building_info
                    for building_info in list_of_building_info
                    if building_info["damage"] > 0
                ]
            ),
            "output_filename": file.name_and_extension(output_filename),
        },
    ):
        return False

    all_area = [building_info["area"] for building_info in list_of_building_info]
    all_area_hist, bin_edges = np.histogram(all_area, bins=10)
    damaged_area_hist = np.histogram(
        [
            building_info["area"]
            for building_info in list_of_building_info
            if building_info["damage"] > 0
        ],
        bins=bin_edges,
    )[0]
    non_damaged_area_hist = np.histogram(
        [
            building_info["area"]
            for building_info in list_of_building_info
            if building_info["damage"] == 0
        ],
        bins=bin_edges,
    )[0]

    plt.figure(figsize=(10, 5))
    bar_width = (bin_edges[1] - bin_edges[0]) / 5
    bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])
    plt.bar(
        bin_centers + bar_width,
        all_area_hist,
        width=bar_width,
        label="All",
        align="center",
        color="black",
        alpha=0.5,
    )
    plt.bar(
        bin_centers + 2 * bar_width,
        damaged_area_hist,
        width=bar_width,
        label="Not Damaged",
        align="center",
        color="green",
        alpha=0.5,
    )
    plt.bar(
        bin_centers + 3 * bar_width,
        non_damaged_area_hist,
        width=bar_width,
        label="Damaged",
        align="center",
        color="red",
        alpha=0.5,
    )
    plt.title("Distribution of Building Areas")
    plt.xlabel("Building Area (sq m)")
    plt.ylabel("Frequency")
    plt.grid(True)
    plt.legend()
    filename = objects.path_of(
        "area-damage-histogram.png",
        object_name,
    )
    if not file.save_fig(filename):
        return False
    if not sign_filename(
        filename=filename,
        header=objects.signature(
            info=reference_filename,
            object_name=datacube_id,
        )
        + [
            f"{object_name}",
            file.name_and_extension(footprint_filename),
            "{} building(s)".format(len(list_of_building_info)),
        ],
        footer=signature(),
    ):
        return False

    plt.figure(figsize=(10, 5))
    colors = [
        "red" if building_info["damage"] > 0 else "green"
        for building_info in list_of_building_info
    ]
    plt.scatter(
        all_area,
        [building_info["damage"] for building_info in list_of_building_info],
        c=colors,
        alpha=0.5,
    )
    plt.title("Scatter Plot of Building Area vs Damage")
    plt.xlabel("Building Area (sq m)")
    plt.ylabel("Damage")
    plt.ylim(0, 1)
    plt.grid(True)
    filename = objects.path_of(
        "area-damage-scatter.png",
        object_name,
    )
    if not file.save_fig(filename):
        return False
    if not sign_filename(
        filename=filename,
        header=objects.signature(
            info=reference_filename,
            object_name=datacube_id,
        )
        + [
            f"{object_name}",
            file.name_and_extension(footprint_filename),
            "{} building(s)".format(len(list_of_building_info)),
        ],
        footer=signature(),
    ):
        return False

    return True



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from palisades import NAME
from palisades import env
from palisades.buildings.analysis import analyze_buildings
from palisades.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="analyze",
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--buffer",
    type=float,
    default=env.PALISADES_DEFAULT_BUFFER_M,
)
parser.add_argument(
    "--verbose",
    type=int,
    default=0,
    help="0|1",
)

args = parser.parse_args()

success = False
if args.task == "analyze":
    success = analyze_buildings(
        object_name=args.object_name,
        buffer=args.buffer,
        verbose=args.verbose == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import List

from blue_options.terminal import show_usage, xtra
from abcli.help.generic import help_functions as generic_help_functions


def analyze_options(
    mono: bool,
    cascade: bool = False,
):
    return "".join(
        ["buffer=<buffer>"]
        + (
            []
            if cascade
            else [
                xtra(",~download,dryrun,~ingest,", mono=mono),
                "upload",
            ]
        )
    )


analyze_details = {
    "buffer: in meters.": [],
}


def help_analyze(
    tokens: List[str],
    mono: bool,
) -> str:

    return show_usage(
        [
            "palisades",
            "buildings",
            "analyze",
            f"[{analyze_options(mono=mono)}]",
            "[.|<object-name>]",
        ],
        "analyze the buildings in <object-name>.",
        analyze_details,
        mono=mono,
    )


def query_options(mono: bool):
    return "".join(
        [
            "country_code=<iso-code>",
            xtra(",country_name=<country-name>,overwrite,", mono=mono),
            "source=<source>",
        ]
    )


query_details = {
    "country-name: for Microsoft, optional, overrides <iso-code>.": [],
    "iso-code: Country Alpha2 ISO code: https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes": [
        "Canada: CA",
        "US: US",
    ],
    "source: microsoft | osm | google": [],
    "calls: https://github.com/microsoft/building-damage-assessment/blob/main/download_building_footprints.py": [],
}


def help_download_footprints(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("~download,dryrun,", mono=mono),
            "filename=<filename>",
            xtra(",upload", mono=mono),
        ]
    )

    return show_usage(
        [
            "palisades",
            "buildings",
            "download_footprints",
            f"[{options}]",
            "[.|<input-object-name>]",
            f"[{query_options(mono=mono)}]",
            "[-|<output-object-name>]",
        ],
        "aoi:<input-object-name>/<filename> -download-building-footprints-> <output-object-name>.",
        query_details,
        mono=mono,
    )


help_functions = {
    "analyze": help_analyze,
    "download_footprints": help_download_footprints,
}



import os

from blue_options.help.functions import get_help
from blue_objects import file, README
from blue_geo import ICON as blue_geo_ICON
from roofai import ICON as roofai_ICON

from palisades.help.functions import help_functions
from palisades import NAME, VERSION, ICON, REPO_NAME, MARQUEE

# refactor

list_of_menu_item = {
    "STAC Catalog: Maxar Open Data": {
        "ICON": blue_geo_ICON,
        "url": "https://github.com/kamangir/blue-geo/tree/main/blue_geo/catalog/maxar_open_data",
        "marquee": "https://github.com/kamangir/assets/blob/main/blue-geo/Maxar-Open-Datacube.png?raw=true",
        "title": '["Satellite imagery for select sudden onset major crisis events"](https://www.maxar.com/open-data/)',
    },
    "Vision Algo: Semantic Segmentation": {
        "ICON": roofai_ICON,
        "url": "https://github.com/kamangir/palisades/blob/main/palisades/docs/step-by-step.md",
        "marquee": "https://github.com/kamangir/assets/raw/main/palisades/prediction-lres.png?raw=true",
        "title": "[segmentation_models.pytorch](https://github.com/qubvel-org/segmentation_models.pytorch)",
    },
    "Analytics: Building Damage": {
        "ICON": ICON,
        "url": "https://github.com/kamangir/palisades/blob/main/palisades/docs/building-analysis.md",
        "marquee": "https://github.com/kamangir/assets/blob/main/palisades/building-analysis-2.png?raw=true",
        "title": "Microsoft, OSM, and Google footprints through [microsoft/building-damage-assessment](https://github.com/microsoft/building-damage-assessment)",
    },
    "template": {
        "ICON": ICON,
        "url": "#",
        "marquee": "",
        "title": "",
    },
}


items = [
    "{}[`{}`]({}) [![image]({})]({}) {}".format(
        menu_item["ICON"],
        menu_item_name,
        menu_item["url"],
        menu_item["marquee"],
        menu_item["url"],
        menu_item["title"],
    )
    for menu_item_name, menu_item in list_of_menu_item.items()
    if menu_item_name != "template"
]


def build():
    return all(
        README.build(
            items=items,
            path=os.path.join(file.path(__file__), path),
            ICON=ICON,
            NAME=NAME,
            help_function=lambda tokens: get_help(
                tokens,
                help_functions,
                mono=True,
            ),
            VERSION=VERSION,
            REPO_NAME=REPO_NAME,
        )
        for items, path in [
            (
                items,
                "..",
            ),
            (
                [],
                "docs/step-by-step.md",
            ),
            (
                [],
                "docs/release-one.md",
            ),
            (
                [],
                "docs/building-analysis.md",
            ),
        ]
    )
